class XlsxOperationException(Exception):
    """
    This exception is raised for errors that occur when processing .xlsx
    spreadsheet files.

    It has no custom implmentation. It is here to provide a class that
    can be specifically caught in an except statement.

    See also: https://docs.python.org/3/library/exceptions.html
    """

    pass
